From 8c0d147e465ca48b8cce8b6cfff2daede1e91add Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Artur=20=C5=9Awigo=C5=84?= <aswigon@aswigon.pl>
Date: Fri, 12 May 2017 20:25:31 +0200
Subject: [v2] wrc: Allow quotes in resource nameID.
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

Fixes https://bugs.winehq.org/show_bug.cgi?id=786

This patch introduces two new parser states to differentiate between strings
and nameIDs. Nested quotes are allowed as long as they appear in pairs.

Compared to the previous patch:
https://bugs.winehq.org/attachment.cgi?id=57805&action=diff
this patch does not introduce a new variable to address the issue of a parse
error if the first nameID in the file is quoted, but uses a marker value in
id_dblquote, because neither setting want_id and wanted_id at file scope nor
setting them in YY_USER_INIT works.

VS2017 rc.exe accepts quotes is IDs that do not necessarily begin with a
quote and has a concept of nesting where the number innermost quotes can be
odd. Reproducing this behavior is beyond the scope of this patch as it would
require substantial changes to wrc.

Tested on macOS 10.12 and Fedora 25.

Signed-off-by: Artur Świgoń <aswigon@aswigon.pl>
---
 tools/wrc/parser.l | 56 ++++++++++++++++++++++++++++++++++++++++++++++++++++--
 1 file changed, 54 insertions(+), 2 deletions(-)

diff --git a/tools/wrc/parser.l b/tools/wrc/parser.l
index 32a231b5a2..b3595208f5 100644
--- a/tools/wrc/parser.l
+++ b/tools/wrc/parser.l
@@ -67,6 +67,10 @@
  *			  escaped '\0'.
  */
 
+/* Exclusive double-quoted nameID handling */
+%x tkid_dbl
+/* Exclusive single-quoted nameID handling */
+%x tkid_sgl
 /* Exclusive string handling */
 %x tkstr
 /* Exclusive unicode string handling */
@@ -138,6 +142,11 @@ static int wbufalloc = 0;
 
 static int current_codepage = -1;  /* use language default */
 
+/* Used for differentiating between single and double-quoted nameID's.
+ * The initial value (-1) is a marker because wanted_id cannot be set at
+ * the beginning. */
+static int id_dblquote = -1;
+
 /*
  * This one is a bit tricky.
  * We set 'want_id' in the parser to get the first
@@ -402,8 +411,31 @@ static unsigned long xstrtoul(const char *nptr, char **endptr, int base)
 0[oO][0-7]+[lL]?	{ parser_lval.num = xstrtoul(yytext+2, 0, 8);
                           return (yytext[yyleng-1] == 'L' || yytext[yyleng-1] == 'l') ? tLNUMBER : tNUMBER; }
 
+<tkid_dbl>[A-Za-z_0-9./\\\"\']+\" |
+<tkid_sgl>[A-Za-z_0-9./\\\"\']+\'	{
+				char *tmp;
+				size_t sz;
+				char q;
+
+				/* Choose the surrounding quote character */
+				q = id_dblquote ? '\"' : '\'';
+				/* Plus 2 - for both the missed quote and a trailing NUL */
+				sz = strlen(yytext) + 2;
+				tmp = xmalloc(sz);
+				/* Put the quote character that was missed */
+				tmp[0] = q;
+				strcpy(&tmp[1], yytext);
+				parser_lval.str = make_string(tmp);
+
+				free(tmp);
+				yy_pop_state();
+				id_dblquote = 0;
+				return tIDENT;
+	}
+
 [A-Za-z_0-9./\\]+	{
 				struct keyword *tok = iskeyword(yytext);
+				id_dblquote = 0;
 
 				if(tok)
 				{
@@ -483,7 +515,17 @@ L\"			{
 	/*
 	 * Normal string scanning
 	 */
-\"			yy_push_state(tkstr); cbufidx = 0;
+\"			{
+				if (wanted_id || id_dblquote == -1)
+				{
+					yy_push_state(tkid_dbl);
+					id_dblquote = 1;
+				}
+				else
+				{
+					yy_push_state(tkstr); cbufidx = 0;
+				}
+			}
 <tkstr>\"{ws}+	|
 <tkstr>\"		{
 				yy_pop_state();
@@ -528,7 +570,17 @@ L\"			{
 	/*
 	 * Raw data scanning
 	 */
-\'			yy_push_state(tkrcd); cbufidx = 0;
+\'			{
+				if (wanted_id || id_dblquote == -1)
+				{
+					yy_push_state(tkid_sgl);
+					id_dblquote = 0;
+				}
+				else
+				{
+					yy_push_state(tkrcd); cbufidx = 0;
+				}
+			}
 <tkrcd>\'		{
 				yy_pop_state();
 				parser_lval.raw = new_raw_data();
-- 
2.13.0

